@@ -1,5 +1,5 @@
 /* openfile.c
- *	Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -39,11 +39,11 @@
 #ifdef USE_ONLY_8DOT3
   #define SKELEXT ".skl"
 #else
-  #define SKELEXT ".skel"
+  #define SKELEXT EXTSEP_S "skel"
 #endif
 
-#ifdef HAVE_DRIVE_LETTERS
-  #define CMP_FILENAME(a,b) stricmp( (a), (b) )
+#if defined (HAVE_DRIVE_LETTERS) || defined (__riscos__)
+  #define CMP_FILENAME(a,b) ascii_strcasecmp( (a), (b) )
 #else
   #define CMP_FILENAME(a,b) strcmp( (a), (b) )
 #endif
@@ -102,15 +102,15 @@ make_outfile_name( const char *iname )
 	return m_strdup("-");
 
     n = strlen(iname);
-    if( n > 4 && (    !CMP_FILENAME(iname+n-4,".gpg")
-		   || !CMP_FILENAME(iname+n-4,".pgp")
-		   || !CMP_FILENAME(iname+n-4,".sig")
-		   || !CMP_FILENAME(iname+n-4,".asc") ) ) {
+    if( n > 4 && (    !CMP_FILENAME(iname+n-4, EXTSEP_S "gpg")
+		   || !CMP_FILENAME(iname+n-4, EXTSEP_S "pgp")
+		   || !CMP_FILENAME(iname+n-4, EXTSEP_S "sig")
+		   || !CMP_FILENAME(iname+n-4, EXTSEP_S "asc") ) ) {
 	char *buf = m_strdup( iname );
 	buf[n-4] = 0;
 	return buf;
     }
-    else if( n > 5 && !CMP_FILENAME(iname+n-5,".sign") ) {
+    else if( n > 5 && !CMP_FILENAME(iname+n-5, EXTSEP_S "sign") ) {
 	char *buf = m_strdup( iname );
 	buf[n-5] = 0;
 	return buf;
@@ -155,6 +155,8 @@ ask_outfile_name( const char *name, size_t namelen )
 	fname = defname; defname = NULL;
     }
     m_free(defname);
+    if (fname)
+        trim_spaces (fname);
     return fname;
 }
 
@@ -215,13 +217,25 @@ open_outfile( const char *iname, int mode, IOBUF *a )
 		strcat( buf, newsfx );
 	  #else
 	    buf = m_alloc(strlen(iname)+4+1);
-	    strcpy(stpcpy(buf,iname), mode==1 ? ".asc" :
-				      mode==2 ? ".sig" : ".gpg");
+	    strcpy(stpcpy(buf,iname), mode==1 ? EXTSEP_S "asc" :
+				      mode==2 ? EXTSEP_S "sig" : EXTSEP_S "gpg");
 	  #endif
 	    name = buf;
 	}
 
-	if( overwrite_filep( name ) ) {
+        rc = 0;
+	while( !overwrite_filep (name) ) {
+            char *tmp = ask_outfile_name (NULL, 0);
+            if ( !tmp || !*tmp ) {
+                m_free (tmp);
+                rc = G10ERR_FILE_EXISTS;
+                break;
+            }
+            m_free (buf);
+            name = buf = tmp;
+        }
+
+	if( !rc ) {
 	    if( !(*a = iobuf_create( name )) ) {
 		log_error(_("%s: can't create: %s\n"), name, strerror(errno) );
 		rc = G10ERR_CREATE_FILE;
@@ -229,8 +243,6 @@ open_outfile( const char *iname, int mode, IOBUF *a )
 	    else if( opt.verbose )
 		log_info(_("writing to `%s'\n"), name );
 	}
-	else
-	    rc = G10ERR_FILE_EXISTS;
 	m_free(buf);
     }
     return rc;
@@ -250,12 +262,12 @@ open_sigfile( const char *iname )
 
     if( iname && !(*iname == '-' && !iname[1]) ) {
 	len = strlen(iname);
-	if( len > 4 && ( !strcmp(iname + len - 4, ".sig")
-                        || ( len > 5 && !strcmp(iname + len - 5, ".sign") )
-                        || !strcmp(iname + len - 4, ".asc")) ) {
+	if( len > 4 && ( !strcmp(iname + len - 4, EXTSEP_S "sig")
+                        || ( len > 5 && !strcmp(iname + len - 5, EXTSEP_S "sign") )
+                        || !strcmp(iname + len - 4, EXTSEP_S "asc")) ) {
 	    char *buf;
 	    buf = m_strdup(iname);
-	    buf[len-4] = 0 ;
+	    buf[len-(buf[len-1]=='n'?5:4)] = 0 ;
 	    a = iobuf_open( buf );
 	    if( a && opt.verbose )
 		log_info(_("assuming signed data in `%s'\n"), buf );
@@ -282,14 +294,14 @@ copy_options_file( const char *destdir )
 	return;
 
     fname = m_alloc( strlen(datadir) + strlen(destdir) + 15 );
-    strcpy(stpcpy(fname, datadir), "/options" SKELEXT );
+    strcpy(stpcpy(fname, datadir), DIRSEP_S "options" SKELEXT );
     src = fopen( fname, "r" );
     if( !src ) {
 	log_error(_("%s: can't open: %s\n"), fname, strerror(errno) );
 	m_free(fname);
 	return;
     }
-    strcpy(stpcpy(fname, destdir), "/options" );
+    strcpy(stpcpy(fname, destdir), DIRSEP_S "options" );
     dst = fopen( fname, "w" );
     if( !dst ) {
 	log_error(_("%s: can't create: %s\n"), fname, strerror(errno) );
@@ -324,7 +336,7 @@ try_make_homedir( const char *fname )
      * To cope with HOME, we do compare only the suffix if we see that
      * the default homedir does start with a tilde.
      */
-    if( opt.dry_run )
+    if( opt.dry_run || opt.no_homedir_creation )
 	return;
 
     if ( ( *defhome == '~'
