@@ -1,5 +1,5 @@
 /* options.h
- *	Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
+ * Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
  *
  * This file is part of GnuPG.
  *
@@ -24,7 +24,16 @@
 
 #undef ENABLE_COMMENT_PACKETS  /* don't create comment packets */
 
+#ifndef EXTERN_UNLESS_MAIN_MODULE
+/* Norcraft can't cope with common symbols */
+ #if defined (__riscos__) && !defined (INCLUDED_BY_MAIN_MODULE)
+  #define EXTERN_UNLESS_MAIN_MODULE extern
+ #else
+  #define EXTERN_UNLESS_MAIN_MODULE 
+ #endif
+#endif
 
+EXTERN_UNLESS_MAIN_MODULE
 struct {
     int verbose;
     int quiet;
@@ -35,24 +44,31 @@ struct {
     int dry_run;
     int list_only;
     int textmode;
+    int expert;
+    int ask_sig_expire;
+    int ask_cert_expire;
     int batch;	    /* run in batch mode */
     int answer_yes; /* answer yes on most questions */
     int answer_no;  /* answer no on most questions */
     int check_sigs; /* check key signatures */
     int with_colons;
     int with_key_data;
+    int with_fingerprint; /* opt --with-fingerprint active */
     int fingerprint; /* list fingerprints */
     int list_sigs;   /* list signatures */
     int no_armor;
-    int list_packets; /* list-packets mode */
+    int list_packets; /* list-packets mode: 1=normal, 2=invoked by command*/
     int def_cipher_algo;
     int force_v3_sigs;
+    int force_v4_certs;
     int force_mdc;
+    int disable_mdc;
     int def_digest_algo;
     int def_compress_algo;
     const char *def_secret_key;
     char *def_recipient;
     int def_recipient_self;
+    int def_check_level;
     int no_comment;
     int no_version;
     int marginals_needed;
@@ -63,6 +79,8 @@ struct {
     int compress_keys;
     int compress_sigs;
     int always_trust;
+    int pgp2;
+    int pgp6;
     int rfc1991;
     int rfc2440;
     int pgp2_workarounds;
@@ -71,17 +89,38 @@ struct {
     const char *set_filename;
     const char *comment_string;
     int throw_keyid;
+    int show_photos;
+    const char *photo_viewer;
     int s2k_mode;
     int s2k_digest_algo;
     int s2k_cipher_algo;
     int not_dash_escaped;
     int escape_from;
     int lock_once;
-    const char *keyserver_name;
+    char *keyserver_scheme;
+    char *keyserver_host;
+    char *keyserver_port;
+    struct
+    {
+      int verbose;
+      int fast_import:1;
+      int include_revoked:1;
+      int include_disabled:1;
+      int use_temp_files:1;
+      int keep_temp_files:1;
+      int refresh_add_fake_v3_keyids:1;
+      STRLIST other;
+    } keyserver_options;
+    int exec_disable;
+    int no_perm_warn;
+    char *temp_dir;
     int no_encrypt_to;
     int interactive;
     STRLIST notation_data;
-    const char *set_policy_url;
+    int show_notation;
+    const char *sig_policy_url;
+    const char *cert_policy_url;
+    int show_policy_url;
     int use_embedded_filename;
     int allow_non_selfsigned_uid;
     int allow_freeform_uid;
@@ -89,17 +128,25 @@ struct {
     ulong set_filesize;
     int honor_http_proxy;
     int fast_list_mode;
+    int fixed_list_mode;
     int ignore_time_conflict;
+    int ignore_valid_from;
     int ignore_crc_error;
     int command_fd;
     int auto_key_retrieve;
     const char *override_session_key;
     int show_session_key;
     int use_agent;
+    const char *gpg_agent_info;
     int merge_only;
-    int allow_secret_key_import;
     int try_all_secrets;
     int no_expensive_trust_checks;
+    int no_sig_cache;
+    int no_sig_create_check;
+    int no_auto_check_trustdb;
+    int preserve_permissions;
+    int no_homedir_creation;
+    int show_keyring;
 } opt;
 
 
